;==============================================================================
; MARK.ASM - mark a position in memory,
;            above which TSRs will later be cleared by RELEASE
;            MARK can be called multiple times, each RELEASE will clear
;            above the last MARK called.
; If MARK is called with something on the command line, that text will
; be stored in the program segment prefix at offset 80H, where it is
; later accessible to RELEASE for a "named" release.
;
; Syntax:    MARK [/Q] [!][MarkName]
;==============================================================================
; written for TASM
; Copyright (c) 1986,1991 Kim Kokkonen, TurboPower Software.
; May be freely distributed but not sold except by permission.
; telephone: 719-260-6641, Compuserve 76004,2611
;==============================================================================
; VERSION 1.0  2/8/86
;     original public release
;     (thanks to Neil Rubenking for an outline of the method used)
; :
; long intervening history
; :
; VERSION 3.0  9/24/91
;   add Quiet option
;   add code for tracking high memory
;==============================================================================

Cseg      segment public para
          assume  cs:Cseg, ds:Cseg, es:nothing

          org     81H
cmdlin    label   byte               ;first character of command line

          org     100H
comentry: jmp     doit

;put the following in the MAP file
public idstr,vector,egasav,intcom,emscnt,emsmap

idstr  db      'M3.0 PARAMETER BLOCK FOLLOWS' ;used to find this TSR

;**** the following will be overwritten by the vector table image *************
;success message and version number
didit  db      'MARK 3.0, Copyright 1991 TurboPower Software',13,10
didit2 db      'Marked current memory position',13,10,36
errst  db      'Invalid command line',13,10,36
;file name for testing EMS presence
emsnm  db      'EMMXXXX0',0
quiet  db      0
;******************************************************************************

vector equ     0120H           ;offset in code seg where vector table is copied
veclen equ     0400H           ;1024 bytes for vectors
egasav equ     vector+veclen
egalen equ     8               ;8 bytes for EGA save area
intcom equ     egasav+egalen
intlen equ     10H             ;16 bytes for interapps communication area
parent equ     intcom+intlen
parlen equ     2               ;2 bytes for parent's PSP
emscnt equ     parent+parlen
emsmap equ     emscnt+2
;mcbcnt and mcbmap actually follow after used portion of emsmap
mcbcnt equ     emsmap+400H     ;count of allocated mcbs
mcbmap equ     mcbcnt+2        ;array of mcb segments
;newloc follows after the longest possible emsmap and mcbmap
newloc equ     mcbmap+400H     ;location of relocated code

;*****************************************************************************

doit   proc    near

;relocate ourselves out of the way
       mov     di,newloc
       push    di                       ;will act as a return address
       mov     si,offset parse
       mov     cx,lastcode-parse
       cld
       rep     movsb
       ret                              ;"return" to the relocated code

;scan command line for /Q option
parse: mov     si,offset cmdlin
       cld
get1:  lodsb
       cmp     al,13                    ;end of command line?
       je      pmess
       cmp     al,'/'                   ;option switch?
       je      getop
       cmp     al,'-'
       jne     get1                     ;loop around
getop: mov     di,si                    ;save option position
       lodsb
       cmp     al,'a'
       jb      noup
       cmp     al,'z'
       ja      noup
       and     al,0DFH                  ;uppercase
noup:  cmp     al,'Q'
       jne     error
       mov     quiet,1                  ;set quiet flag
       dec     di
       mov     al,' '
       stosb                            ;clear option
       stosb
       jmp     get1                     ;loop around

error: mov     dx,offset errst
       mov     ah,9
       int     21H
       mov     ax,4C01H
       int     21H

;print message if not quiet
pmess: cmp     quiet,0
       jnz     chkems
       mov     dx,offset didit
       mov     ah,9
       int     21H                      ;write success message

;determine whether EMS is present
chkems:mov     dx,offset emsnm
       mov     ax,3D00H
       int     21H
       jnc     emspres                  ;EMS driver installed
       xor     bx,bx
       jmp     short stocnt

;EMS present, close the open handle first
emspres:
       mov     bx,ax
       mov     ah,3EH
       int     21H

;store the EMS page map
       mov     ah,4DH
       mov     di,emsmap
       xor     bx,bx                    ;required by some versions of EMM
       cld                              ;required by some versions of EMM
       int     67H

;store the number of active EMS handles
stocnt:
       mov     ds:[emscnt],bx           ;count of active handles

;store the interrupt vector table (overwrites initial messages)
       xor     ax,ax
       mov     ds,ax                    ;source address segment 0
       assume  ds:nothing
       xor     si,si                    ;offset 0
       mov     di,vector                ;destination offset, es=cs already
       mov     cx,veclen shr 1          ;count of words to store
       rep     movsw                    ;copy vectors to our table

;store the EGA save area
       mov     ax,40H
       mov     ds,ax                    ;point to BIOS data area
       mov     si,00A8H                 ;EGA save area pointer
       mov     di,egasav
       mov     cx,egalen shr 1
       rep     movsw                    ;copy information to our save area

;store the interapplications communication area
       mov     si,00F0H                 ;interapps communication area address
       mov     di,intcom
       mov     cx,intlen shr 1
       rep     movsw                    ;copy information to our save area

;store the parent's PSP segment
       mov     ax,cs                    ;CS = PSP
       mov     ds,ax                    ;now DS = PSP
       mov     ax,ds:[16h]              ;get parent's PSP from our PSP
       mov     ds:[parent],ax           ;store in data save area

;store the mcb chain
       mov     ax,5802H
       int     21H                      ;get UMB link status
       jnc     linkok
       mov     al,0
linkok:xor     ah,ah
       push    ax                       ;save link status
       mov     ax,5803H
       mov     bx,1                     ;open up high memory
       int     21H

       mov     ah,52H                   ;get first mcb segment
       int     21H
       mov     ax,es:[bx-2]             ;ax=first mcb
       push    cs
       pop     es                       ;es=cs

       mov     di,cs:[emscnt]           ;get starting address of mcbmap
       shl     di,1
       shl     di,1
       add     di,emsmap
       mov     si,di                    ;cs:[si] -> mcbcnt
       add     di,2                     ;cs:[di] -> mcbmap
       xor     cx,cx                    ;cx will count mcbs
       cld
       push    ds

mcbnext:
       stosw                            ;store mcb segment held by ax
       mov     ds,ax                    ;ds:[0] points to mcb
       mov     ax,ds:[1]                ;get mcb owner
       stosw                            ;store it
       inc     cx                       ;increment count
       cmp     byte ptr ds:[0],'Z'      ;end of mcb chain?
       je      mcbdone
       mov     ax,ds                    ;restore ax to mcb segment
       inc     ax                       ;skip over mcb itself
       add     ax,ds:[3]                ;add length of memory block
       jmp     mcbnext

mcbdone:
       pop     ds
       mov     ax,5803H
       pop     bx                       ;restore high memory link
       int     21H

;terminate and stay resident
gores: mov     [si],cx                  ;store mcbcnt
       mov     dx,di                    ;dx = past last used mcb segment
       add     dx,000FH                 ;add paragraph roundup
       mov     cl,4
       shr     dx,cl                    ;convert to paragraphs
       mov     ax,3100H
       int     21H                      ;terminate but stay resident

lastcode:
doit    endp

Cseg    ends
        end     ComEntry
